// PIC18F242 Boot Loader
//
// Code size = 0x474 with all comments removed
//           = 0x665 with comments
//
// PICLITE compilation
//


/*
 * 18Fxxx BOOT LOADER Copyright (C) 2007 HI-TECH Software
 *	This software is freely distributable and may be used
 *	for any purpose. No warranty of any kind is provided
 *	and all use is entirely at your own risk.
 *	
 */
#include <PIC18.h>
#include "bootldr.h"

#if defined(MPLAB_ICD)
 __CONFIG(4,DEBUGEN & LVPDIS);	// required ICD2 settings
#endif

#if (PROG_START != BOOT_BLOCK_SIZE) || defined(MPLAB_ICD)
 __CONFIG(6,UNPROTECT);	// bootloader protected by address rejection
 #define _WPB_IS_OFF	1
#else
 __CONFIG(6,WPB & UNPROTECT);	// bootloader protected by WPB
#endif

//#if (_ERRATA_TYPES & ERRATA_4000)
//  #error Command line option --errata=default,-4000 must be specified when compiling this code for this device.
//#endif

/* Function prototypes */
void checksum(void);
void putch(unsigned char);
void message(near char *s);
unsigned char gx(void);
unsigned char g2x(void);
void zap(void);
void table_write(void);
void flash8(void);
void clear_buffer(void);

persistent near unsigned char buff[8];	/* Internal buffer used for writing to program memory */
persistent near unsigned char DO_NOT_INCREMENT;
persistent near unsigned char cksum,rectype,bcount;
persistent near unsigned char delay_time;
persistent near unsigned char index;
persistent near unsigned short erase;

#ifdef VERBOSE
near unsigned char START_MSG[] = "HI-TECH Software (C)2008\n";
near unsigned char DOWNLOAD_MSG[] = "\rDownload-";


#endif

//const unsigned char MSG[12] = "PIC18F242  ";


#asm
psect intcode	// redirect interrupts to new routine addresses
	goto	0x308
psect intcodelo // redirect low priority interrupts to new routine addresses
    goto    0x318
#endasm

/* putch() - outputs 1 byte to the serial port */
void
putch(unsigned char byte){
	while(!TRMT);	/* set when register is empty */
	TXREG = byte;	/* output one byte */
}



/* check the checksum 
 * NOTE: if not in verbose mode, the code will use the even more compact
 * assembler version of the checksum() function defined above.
 */

void checksum(void){
	g2x();
	if(cksum)	// if checksum does not add to zero, bad check, reset
		RESET();
	putch('\r');	// echo each hex record on a new line
}
	

/* Get a 4 bit HEX number from the serial port */
unsigned char gx(void){
	while(!RCIF);
	EEDATA=RCREG;
	if(EEDATA>='A')
		return ((EEDATA - (unsigned char)'A')+10);
	return (EEDATA - '0');
}

/* recieve a hexadecimal byte, add it to the checksum */
unsigned char g2x(void)
{
	unsigned char temp = (gx()<<4);
	temp += gx();
	cksum+=temp;
	return temp;
}

/* Initiate a write to memory */
void
zap(void){
	WREN=1;
	EECON2=0x55;
	EECON2=0xAA;
	WR=1;
	NOP();
	while(WR);
	WREN=0;
}

/* transfer a byte from RAM to internal flash register(s) */
void table_write(void){
	if(DO_NOT_INCREMENT)	// address is already loaded in TBLPTR, no pre-increment needed
		asm("tblwt*");
	else
		asm("tblwt+*");	// otherwise TBLPTR must be pre-incremented
	DO_NOT_INCREMENT=0;
}

/* Commit the internal buffer to flash */
void flash8(void){
	if(DO_NOT_INCREMENT)
		TBLPTRL&= ~(8-1);	// point to start of flash panel
	for(index=0;index<8;)
	{	
		TABLAT = buff[index++];
		table_write();
	}
	zap();
}

void clear_buffer(void){
	unsigned char x;
	for(x=sizeof(buff);x--;){
		buff[x]=FILL_BYTE;	// initialize buffer with known values
	}
}


void SCI_CRLF(void)
{
   while(!TRMT);  TXREG = 13;
   while(!TRMT);  TXREG = 10;
}

void SCI_Send(unsigned char X)
{
   while(!TRMT);  TXREG = X;
}
      


//  **************  Main Routine  *************************

void main(void)
{
   unsigned int i, j, TIME;
   unsigned char FLAG;

// Modified code for 115kpbs
//   TXSTA = 0x22;           // 8-bit no parity
//   RCSTA = 0x90;           // continuous receive
//   BRGH  = 1;
//   SPBRG = 10;             // 115,200 BPS
//   SYNC  = 0;

// 9600 bps communications
	SPBRG = 31;		
	TXSTA = 0x22;		
	RCSTA = 0x90;		
	TRISC=0x80;	

//   init_comms();	// sets up the serial port for communication

// Welcome Message
//   SCI_CRLF();
//   for (i=0; i<10; i++) SCI_Send(MSG[i]);
//   SCI_CRLF();
//   for (i=0; i<17; i++) SCI_Send(MSG1[i]);
   SCI_Send('>');
   SCI_CRLF();

   TIME = 4;
   FLAG = 0;
   RCIF = 0;
   j = 100;

   while( (TIME > 0) && (RCREG != 10) && (RCREG != 13) ) {
      if (j>0) j = j - 1;
      if (j == 0) {
         TIME -= 1;
         SCI_Send(TIME + '0');
         j = 300;
      }
      for (i=0; i<1000; i++);
   }
   if (TIME == 0) {
      SCI_Send(':');
      T1CON = 0;
      TXSTA = 0;
      RCSTA = 0;
      asm("goto 0x300");
      }

// else, you're downloading a hex file


//   SCI_CRLF();
//     SCI_Send('d');
//   SCI_CRLF();
   
  
	TXREG=':';	// the prompt to invite you to download a new program

	TBLPTRU=0;
	erase=PROG_START;
	/* Clear program memory prior to program download */
	while(1)
	{
		TBLPTRL=(unsigned char)erase;
		TBLPTRH=(unsigned char)(erase>>8);
		EECON1=0x90;	// set up for flash erasure
		zap();
		erase+=_FLASH_ERASE_SIZE;
#if defined(UPPER_ADDRESS_BYTE)	// cases where top of memory greater than 10000h
		if(erase==0)
			TBLPTRU++;
		if(TBLPTRU==UPPER_ADDRESS_BYTE)	// continues to next "if" or "break"
#endif
#if !defined(UPPER_ADDRESS_BYTE) || (MEM_TOP!=0)
		if(erase==MEM_TOP)	// for MEM_TOPs like 20000h, don't need to test lower 0000h
#endif
			break;
	}

#ifdef UPPER_ADDRESS_BYTE
	TBLPTRU=0;
#endif




/* receive a hex file via the serial port and write it to program memory */
	for(;;)		// loop until end of file
	{
		while (RCREG!=':');		// wait for start of hex file line 

		cksum = bcount = g2x();		// get the byte count 
#if _EEPROMSIZE > 256
		EEADRH = TBLPTRH = g2x();		// get the address
#else
		TBLPTRH = g2x();
#endif
		TBLPTRL = EEADR = g2x();
		
		DO_NOT_INCREMENT = 1;
		rectype = g2x();	// get the record type 
		
		switch(rectype)
		{
			case DATA:	// data record
#if _WPB_IS_OFF
				// to protect bootloader from being overwritten
				if( (FLASH) && (TBLPTRU==0) && (TBLPTRH < (unsigned char)(PROG_START>>8)) )	// to protect bootloader from being overwritten
					break;		// if bootloader address range is threatened, skip this record
#endif
			clear_buffer();
				while(bcount--)
				{	
					TABLAT = EEDATA = buff[(EEADR&(8-1))] = g2x();	// get the data

					if((CONFIG)||(EEPROM))
					{
						if(CONFIG)	// EEPROM/config. bytes are written one byte at a time
							table_write();
						zap();
					}
					else
						if((EEADR & (8-1)) == (8-1))	// program/IDLOCs are flashed x bytes at a time
						{
							flash8();
							clear_buffer();
						}
					EEADR++;
				}	
				if(((EEADR&(8-1))!=0)&&(FLASH))
                                        flash8();

				checksum();
				break;

			case END:	// end of hex file
				checksum();
				TXREG=')';
                while(!TRMT);
                T1CON = 0;
                TXSTA = 0;
                RCSTA = 0;
        		asm("goto 0x300");
				break;

			case EXTEND_ADDRESS:	// extended address record
				while(bcount--)
				{
					EEADR=g2x();	// this byte determines whether EE, Config or ID data
				}
				
				EEPGD=1;
				if(EEADR==0xF0)
					EEPGD=0;	// select for EEPROM
					
				CFGS=0;
				if((EEADR&0xF0)==0x30)
					CFGS=1;		// select for config write
					
				TBLPTRU=EEADR;
				checksum();

				break;
		}
	}
}

